## Copyright (c) 2020
## Global Climate Forum e.V.
## http://www.globalclimateforum.org

## Authors:
##     Steffen Fuerst

## This file is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## This file is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## version 1.01

if (!require("tidyverse")) install.packages("tidyverse")
library(tidyverse)
library(magrittr)

## Implementation
##    The following is a proof of concept implementation, in which the
##    parallelism is simulated by creating a separate list for each
##    simulated PE. The lists are stored in the variable
##    ~modelStateDistributed~. The parameters of the transition functions
##    only include a subset of ~modelStateDistributed[[i]]~, this ensures
##    that the agent PE does not access the "memory" of another PE.

##    We start with the helper function ~mapAgentsToPEs~, that takes a
##    non-distributed model state and creates a distributed version. 


mapAgentsToPEs <- function(modelState, agentTypes, numPEs) {
  edgeTypes <- modelState %>% names %>% discard(. %in% agentTypes)

  ## Store which agent was moved to which PE
  PElist <- tibble(ID = character(), 
                   PE = integer())

  ## Prepare the distributed data structures (the indices are the different PEs)
  modelStateDistributed <- vector("list", numPEs)
  PElists <- vector("list", numPEs)
  for (i in 1:numPEs) {
    PElists[[i]] <- tibble(ID = character(), 
                           PE = integer())
  }

  ## Distribute the agents
  for (a in agentTypes) {
    ## We group the agents by their rowid, so that each PE gets the 
    ## same number (plus/minus one) of agents
    modelState[[a]] %<>% rowid_to_column("PE") %>%
      mutate(PE = ((PE - 1) %% numPEs) + 1) 
    PElist %<>% bind_rows(modelState[[a]]) %>% select(ID, PE)
    perPE <- modelState[[a]] %>% split(.$PE)

    for (i in 1:numPEs) {
      modelStateDistributed[[i]][[a]] <- perPE[[i]] 
      PElists[[i]] %<>% bind_rows(modelStateDistributed[[i]][[a]]) %>%
        select(ID, PE)
    }
  }

  ## Distribute the edges
  for (e in edgeTypes) {
    if (nrow(modelState[[e]]) > 0) {
      for (i in 1:numPEs) {
        ## Find for each PE the edges where an agent of this PE is involved ..
        fID <- modelState[[e]] %>% filter(fromID %in% PElists[[i]]$ID)
        tID <- modelState[[e]] %>% filter(toID %in% PElists[[i]]$ID)
        ## ... and add those edges to the distributed state
        modelStateDistributed[[i]][[e]] <- bind_rows(fID, tID) %>% unique
        ## Also add the PEs of the agents on the other edge pos to the PElists
        PElists[[i]] %<>% bind_rows(fID %>% transmute(ID = toID) %>% left_join(PElist, by = "ID")) %>% unique
        PElists[[i]] %<>% bind_rows(tID %>% transmute(ID = fromID) %>% left_join(PElist, by = "ID")) %>% unique
      }
    }
  }

  list(modelStateDistributed, PElists)
}



## The ~enhanceAgentStates~ function ensures that the state of agents
## at the tail position of the network associated with the next
## transition function is available on the PEs of the agents at the
## head position of the edges. This function is called before the
## transition function itself, which then gets the returned extendend
## model state.


enhanceAgentStates <- function(modelStateDistributed, PElists, agentTypes, edgeType) {
  modelStateEnhanced <- modelStateDistributed

  numPEs <- modelStateDistributed %>% length

  for (i in 1:numPEs) {
    ## We allow the creation of Networks while the simulation is running
    if (! is.null(modelStateDistributed[[i]][[edgeType]]) &&
        nrow(modelStateDistributed[[i]][[edgeType]]) > 0) {
      ## agentPEmap is a tibble(fromID, toID, PE of toID agent)  
      agentPEmap <- modelStateDistributed[[i]][[edgeType]] %>%
        select(fromID, toID) %>%
        left_join(PElists[[i]], by = c("toID" = "ID")) %>%
        unique

      for (a in agentTypes) {
        ## iterate over the other PEs
        for (j in (1:numPEs %>% discard(. == i))) {
          ## find the agents that must be send to other PEs ...
          agentsIDtoSend <- agentPEmap %>% filter(PE == j) %>% select(fromID)
          agentsToSend <- agentsIDtoSend %>% 
            inner_join(modelStateDistributed[[i]][[a]], by = c("fromID" = "ID")) %>%
            unique %>%
            rename(ID = fromID)
          ## ... and add them to the state of this other PE  
          modelStateEnhanced[[j]][[a]] %<>% bind_rows(agentsToSend)
        }
      }
    }
  }

  modelStateEnhanced
}




## ~distributeNetworkState~ takes the result of a transition function,
## and sends the edges to the PEs of the agents on the head or tail position.


distributeNetworkState <- function(modelStateDistributed, PElists, edgeTypes) {
  modelStateEnhanced <- modelStateDistributed

  numPEs <- modelStateDistributed %>% length
  ## create new empty lists
  newPElists <- PElists %>% map(~ .x[0,])

  for (i in 1:numPEs) {
    for (e in edgeTypes) {
      ## We allow the creation of Networks while the simulation is running
      if (! is.null(modelStateDistributed[[i]][[e]]) &&
          nrow(modelStateDistributed[[i]][[e]]) > 0) {
        ## Create two version of the edge tibble, ~matchFrom~ is extended with 
        ## the PEs of the tail, ~matchTo~ with the PEs of the head agents
        matchFrom <- inner_join(modelStateDistributed[[i]][[e]],
                                PElists[[i]], by = c("fromID" = "ID"))
        matchTo <- inner_join(modelStateDistributed[[i]][[e]],
                              PElists[[i]], by = c("toID" = "ID"))
        ## sendNetworkTable is a tibble: fromID toID PE
        sendNetworkTable <- bind_rows(matchFrom, matchTo)
        ## sendPElist is a tibble: ID, PE, toPE
        sendPElist <- bind_rows(sendNetworkTable %>% select(ID = fromID,
                                                            toPE = PE),
                                sendNetworkTable %>% select(ID = toID,
                                                            toPE = PE)) %>%
          inner_join(PElists[[i]], by = "ID") 

        ## Use the send... tibbles to distribute the edges to the PEs
        ## and also to send the (agentID, PE) information to those PEs
        for (j in 1:numPEs) {
          if (i != j) {
            modelStateEnhanced[[j]][[e]] %<>%
              bind_rows(sendNetworkTable %>% filter(PE == j) %>% select(-PE))
          }
          newPElists[[j]] %<>%
            bind_rows(sendPElist %>% filter(toPE == j) %>% select(-toPE)) %>%
            unique
        }
      }
    }
  }

  list(modelStateEnhanced, newPElists)
}



## ~simulate~ combine all this pieces together. ~transform~
## \codemarker{1} takes one agent type and iterate over the agents of
## this type, calling for each agent one of the transition
## functions. Inside of ~transform~ we use the ~limitState~ \codemarker{2} function to
## ensure, that only the state of agents in the tail position of the
## associated network is accessable.

## ~consolidate~ \codemarker{3} gets the collected output of
## ~transform~ from all agents in the system, so ~newModelState~ is a
## list (with the agent types as keys) of lists (with the agents as
## keys) of tables (the tables returned by a transition function), and
## reduces this to a single table per agent type, our new state
## $\Theta_{t+1}$.



simulate <- function(agentTypes, modelParams, 
                     stateInitFunc, observerInitFunc, observerFunc,
                     tFuncs) { 
  ## the edgeTypes are the associated network of the transition functions                   
  edgeTypes <- tFuncs %>% map(~ .x[[2]]) 

  numPEs <- modelParams$numPEs

  transform <- function(type, agents, modelState, tFunc, edgeType) { # |\codemarker{1}|
    limitState <- function(agent) { # |\codemarker{2}|
      incoming <- modelState[[edgeType]] %>% filter(toID == agent$ID)

      ## reduce the agent states to the agents on the networks tail position
      agentTables <- agentTypes %>%
        map(~ modelState[[.x]] %>% filter(ID %in% incoming$fromID)) %>%
        setNames(agentTypes)

      ## reduce the network to edges where the agent is on tail or head position
      edgeTables <- edgeTypes %>% 
        map(~ modelState[[.x]] %>% 
              (function(x) { if(is.null(x) || nrow(x) == 0) NULL else 
                           { x %>% filter(agent$ID == fromID | 
                                          agent$ID == toID)}})) %>%
        setNames(edgeTypes)

      c(agentTables, edgeTables)
    }

    agents %>% transpose %>% map(~ tFunc(as_tibble(.), type, limitState(.x))) 
  }

  consolidate <- function(newModelState) { # |\codemarker{3}|
    flattenLists <- newModelState %>% flatten

    c(agentTypes, edgeTypes) %>% 
      map(~ { flattenLists %>% map(.x) %>% bind_rows }) %>%
      setNames(c(agentTypes, edgeTypes))
  }

  ## create the initial distributed model state and PElist
  modelStateFull <- stateInitFunc(modelParams)
  distributed <- mapAgentsToPEs(modelStateFull, agentTypes, numPEs)
  modelStateDistributed <- distributed[[1]]
  PElist <- distributed[[2]]

  observerState <- observerInitFunc(modelParams)

  ## run the simulation
  for (. in 1:modelParams$numSteps) {
    for (s in tFuncs) { 
      enhancedModelStateD <- enhanceAgentStates(modelStateDistributed, PElist, agentTypes, s[[2]])
      for (i in 1:numPEs) {
        modelStateDistributed[[i]] <- agentTypes %>%
          map(~ transform(.x, modelStateDistributed[[i]][[.x]], enhancedModelStateD[[i]], s[[1]], s[[2]])) %>%
            consolidate() 
      }
      d <- modelStateDistributed %>% distributeNetworkState(PElist, edgeTypes)
      modelStateDistributed <- d[[1]]
      PElist <- d[[2]]
    }
    observerState <- observerFunc(observerState, modelStateDistributed) 
  }

  observerState
}

####################
## Model Example  
####################

## Init Functions

initAgents <- function(modelParams) {
  nb <- modelParams$numBuyer
  ns <- modelParams$numSeller
  list(
    "BUYER" = tibble(
      alpha = runif(nb),
      B = runif(nb, max = 100) %>% ceiling,
      x = 0,
      y = 0,
      ID = paste0("BUYER", 1:nb)
    ),
    "SELLER" = tibble(
      p = runif(ns, min = 0.5, max = 1.5),
      ID = paste0("SELLER", 1:ns)
    )
  )
}

initNetwork <- function(modelParams, modelState) {
  sampleSellers <- function() {
    sample_n(modelState$SELLER, modelParams$knownSellers) %>% pull(ID)
  }

  list("KNOWN_SELLERS" = 
         modelState$BUYER$ID %>%
         map(~ tibble(fromID = sampleSellers(), 
                      toID = .x)) %>%
         bind_rows)
}

initState <- function(modelParams) {
  agentState <- initAgents(modelParams) 
  c(agentState, initNetwork(modelParams, agentState))
}

initObserver <- function(modelParams) {
  tibble(
    sum_x = numeric(),
    sum_y = numeric()
  )
}


## The ~BOUGHT_FROM~ network will be available after calcDemand,
## because every buyer returns an edge to the seller from whom he bought
## the goods.  The ~KNOWN_SELLER~ network is constant, but must be
## recreated in this proof of concept through each transition
## function. However, this is easily done by making each agent return all
## edges where itself is at the head. The sum of these edges then
## results in the ~KNOWN_SELLER~ network.

## The seller does nothing else but return its state. Otherwise the model
## state after calling the function ~calcDemand~ would be without any
## seller.


calcDemand <- function(agent, type, modelState) {
  if (type == "BUYER") {
    seller <- modelState$SELLER %>% sample_n(1)
  }
  switch(type,
         "BUYER" = {
           agent$x <- agent$B * agent$alpha
           agent$y <- agent$B * (1 - agent$alpha) / seller$p
           list("BUYER" = agent,
                "BOUGHT_FROM" = tibble(fromID = agent$ID, toID = seller$ID),
                "KNOWN_SELLERS" = modelState$KNOWN_SELLERS %>% filter(toID == agent$ID))
         },
         "SELLER" = {
           list("SELLER" = agent)
         })
}



## In the transition function ~calcPrice~, the network ~KNOWN_SELLER~
## can be treated as in ~calcDemand~, and the network ~BOUGHT_FROM~ can
## be ignored, since it is rebuilt in the next ~calcDemand~ call.


calcPrice <- function(agent, type, modelState) {
  quant <- modelState$BUYER %>% select(x, y) %>% summarise_all(sum)
  switch(type,
         "BUYER" = 
           list("BUYER" = agent,
                "KNOWN_SELLERS" = modelState$KNOWN_SELLERS %>% filter(toID == agent$ID)
                ),
         "SELLER" = {
           agent$p <- if (quant$x > 0) { quant$y / quant$x * agent$p } else { agent$p }
           list("SELLER" = agent)
         }
         )
}



## The framework does the parallelization mostly automatically, it is not
## necessary to keep the parallelization in mind while writing the
## transition functions or the initialization of the model state remains the
## same. 

## The observer on a PE, however, can only observe the part of the
## model state which is exactly on this PE, which must be considered
## by the modeller. The framework to be developed, however, will
## provide functions that support him, similar to the following sumD
## function.


observeDistributed <- function(observerState, modelStateDistributed) {
  sumD <- function(type, var) {
    modelStateDistributed %>% map(type) %>% map(var) %>% unlist %>% sum
  }

  observerState %>% add_row(
                 sum_x = sumD("BUYER", "x"),
                 sum_y = sumD("BUYER", "y")
               )
}



## Finally, here is a run of this example model. The result of the
## simulated function is a table with the total demand for x and y for
## each time step. We plot a time series of the difference of this
## demand, which in equilibrium should be exactly 0.



plot <- simulate(agentTypes = c("BUYER", "SELLER"),
                 modelParams = tibble(numBuyer = 20, numSeller = 5, knownSellers = 2,
                                      numSteps = 10, numPEs = 5),
                 stateInitFunc = initState, 
                 observerInitFunc = initObserver,
                 observerFunc = observeDistributed,
                 tFuncs = list(c(calcDemand, "KNOWN_SELLERS"), 
                               c(calcPrice, "BOUGHT_FROM"))
                ) %>%
            rowid_to_column("step") %>%
            ggplot(aes(x = step, y = sum_x - sum_y)) + geom_line()

print(plot)

